/*
 * Copyright (c) 2007 Sun Microsystems, Inc. All rights reserved
 * Use is subject to license terms.
 *
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * http_atom.c: Minimal implementation of the Atom Publishing Protocol, which
 *  also involves generating Atom feeds.
 *
 * - Tim Bray (started 2007/04/26)
 */

/*
 * For this first cut, we're all convention over configuration, so there's
 *  just one one directive, AtomPub:
 *
 * AtomPub /public/uri/path /real/directory/where/data/lives "Title" "Author"
 *
 * Let's call the base directory D; it has a subtree like this:
 *
 *  D/
 *  D/pub/ human-oriented feeds and entries and so on go here.
 *  D/atom/ publishing-protocol gunk goes here.
 * 
 *  D/atom/service the service document.
 *
 * The directory structure distinguishes between entry and media collections;
 *  to start with, we'll just have two named "entry" and "media".
 * For each feed F:
 *
 * D/pub/F/feed - its feed.
 * D/pub/F/index.html - a front page for the collection.
 * D/pub/F/YYYY/MM/DD/<name> - each file contains what's inside
 *                             the posted atom entry's <content> element.
 *
 * D/atom/e/ - entry collections live here.
 * D/atom/m/ - media collections live here.
 *
 * Let's assume Fe and Fm are entry and media collections respectively.
 *
 * D/atom/e/Fe/
 * D/atom/e/Fe/collection - the URI of the collection you read, or post to.
 * D/atom/e/Fe/YYYY/MM/DD/<name> - the atom entries in the collection,
 *                                 1 per file.
 *
 * D/atom/m/Fm/
 * D/atom/m/Fm/collection - the URI of the collection you read, or post to.
 * D/atom/m/Fm/YYYY/MM/DD/<name> - the media link entries.
 * 
 * D/atom/m/Fm/bits - root of tree for the media resources.
 * D/atom/m/Fm/bits/YYYY/MM/DD/<name> - the media resources.
 *
 * Lots of pieces of code below are extremely hard-wired to this layout,
 *  and changing it will require a lot of re-thinking.
 * There's one more rule.  No directory nor entry filename ever contains a
 *  dot ('.') character; files with dots in their names are used for various
 *  kinds of signaling and out-of-band information.
 */

/*
 * Module workflow (*h means routine name is same as hook name)
 * 1. init phase
 *  "pre_config" (*h), alloc the pubs array
 *  "atompub_directive" sticks the basics about a pub in the pubs array
 *  "post_config" (*h) validates the pubs that were declared and builds
 *  the pubs index.
 * 2. runtime.  Everything routes through the find_pub routine, which looks
 *  up a URI and returns our pub_t record or NULL. We have three runtime
 *  hooks in:
 *  "map_to_storage" (*h) checks to see if the request->uri belongs to
 *    one of our pubs.  If so, further checks to see if it can simply be
 *    served statically as a file, in which case it fills in request->filename.
 *  "type_checker" (*h) figures out our type; since we don't use file
 *    extensions, we have to fill it in manually.
 *  "handler" (*h) takes care of the actual heavy lifting in the POST and
 *    PUT methods.
 */

#include "mod_atom.h"

/* a publication, corresponding to an AtomPub directive in config */
typedef struct {
  /* from config file */
  char * path;    /* the path used by the world to refer to this pub */
  char * dir;     /* the directory where the pub's data is rooted */
  char * title;   /* the pub's human-readable title */
  char * author;  

  /* to help categorize URIs */
  ap_regex_t * html_re;
  ap_regex_t * feed_re;
  ap_regex_t * entry_re;
  ap_regex_t * collection_re;
  ap_regex_t * media_re;
  
  char * service_doc;
  char * entries_prefix;
  char * media_prefix;

  /* Bits->MLE uri helper */
  ap_regex_t * mle_maker;
} pub_t;

/* Package of information about an Atom entry.  It gets written and read in
 *  a fairly ad-hoc fashion by lots of routines.  This is probably not
 *  strictly necessary and arguably represents unnecessary data coupling.
 *  However, it does have the benefit of reducing almost all calls to just
 *  a couple of arguments, makign things easier to read.
 */
typedef struct {
  apr_xml_doc * xml; 
  char *        filename;
  char *        html_name;
  char *        uri;
  pub_t *       pub;
  char *        coll_base;
  char *        id;
} entry_t;

/* Package of information about a media resource.  Like entry_t, this is
 *  a convenience thing and may constitute unnecessary data couling.
 */
typedef struct {
  char * bits_path;
  char * mle_path;
  char * filename;
  int size;
  entry_t * mle; /* media link entry */
} media_t;

/* forward */
/* 11 args, Major refactoring alert, don't ya think? */
static char * generate_collection(char * dirname, apr_pool_t * pool,
				  server_rec * s,
				  pub_t * pub,
				  genxWriter gw,
				  genxSender * sender,
				  int start,
				  int count,
				  genxElement * elements,
				  genxAttribute * attributes,
				  genxNamespace * namespaces);
static int find_num(char * * a_p);
static int replace_media(request_rec * r, media_t * media);
static int replace_mle(request_rec * r, media_t * media);
static void insert_attr(request_rec * r, apr_xml_elem * e, 
			char * name, char * value);
static char * insert_text_content(request_rec * r, apr_xml_elem * e, char * t);
static apr_xml_elem * insert_simple_child(request_rec * r,
					  apr_xml_elem * parent,
					  int ns_ind, char * type,
					  char * content);
static int save_new_media(request_rec * r, media_t * media);
static char * find_coll_base(request_rec * r, pub_t * pub);
static char * yyyymmdd_at(apr_time_t when, apr_pool_t * pool);
static char * write_entry(request_rec * r, apr_xml_doc * doc, char * filename);
static char * prepare_entry(request_rec * r, entry_t * entry);
static int save_new_entry(request_rec * r, entry_t * entry);
static int post_entry(request_rec * r, entry_t * entry);
static int post_media(request_rec * r, pub_t * pub);
static int put_entry(request_rec * r, pub_t * pub, entry_t * entry);
static int put_media(request_rec * r, pub_t * pub, entry_t * entry);
static char * is_prefix(char * s, char * prefix);
static pub_t * find_pub(char * uri);
static char * build_pub(pub_t * pub, server_rec * s, apr_pool_t * pperm, apr_pool_t * ptemp);
static int index_step(int * state, int * mask, int prefix_length,
		      apr_pool_t * pperm, apr_pool_t * pscratch);
static int build_index(apr_pool_t * pperm, apr_pool_t * ptemp) ;
static int get(request_rec * r, pub_t * pub);
static int post(request_rec * r, pub_t * pub);
static int put(request_rec * r, pub_t * pub);
static int delete(request_rec * r, pub_t * pub);
static char * atom_dir(char * name, apr_pool_t * p);
static char * decode_slug(char * slug, apr_pool_t * p);
static char * yyyymmdd_at(apr_time_t when, apr_pool_t * pool);
static char * to_file(char * filename, char * data, apr_pool_t * pool);
static char * write_collection(apr_pool_t * pool, server_rec * s, pub_t * pub, char * name, int size);
static char * atom_time_at(apr_time_t when, apr_pool_t * pool);

/* our globals */
/* These should be OK regardless of what sort of process/thread model we're
 *  using.  Everything is allocated at config time & thus should be
 *  inherited by subsequent code.  At request time, it's all 100% read-only,
 *  these are just lookup indices.
 */
static apr_array_header_t * atom_pubs = NULL;  /* list of known pubs */
static apr_array_header_t * pub_index = NULL;  /* index for mapping uri->pub */
static genxWriter char_util;                   /* for char classes */

/******** Config Time **************/

/* init our 'pubs' global index */
static int pre_config(apr_pool_t * p_conf, apr_pool_t * p_log,
		      apr_pool_t * p_temp)
{
  atom_pubs = apr_array_make(p_conf, 10, sizeof(pub_t));
  return OK;
}

/* Config has been processed.  Initialize the world. */
static int post_config(apr_pool_t * pconf, apr_pool_t * plog,
		       apr_pool_t * ptemp, server_rec * s) 
{
  int i;
  pub_t * publist; 
  char * problem;

  /* build the pub-lookup index */
  publist = (pub_t *) atom_pubs->elts;
  if (build_index(pconf, ptemp)) {
    /* something went wrong */
    return DECLINED;
  }

  /* now we'll build some per-pub data structures */
  for (i = 0; i < atom_pubs->nelts; i++) {
    pub_t * pub = publist + i;

    /* make sure this pub exists */
    if ((problem = build_pub(pub, s, pconf, ptemp))) {
      ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
		   "Can't initialize publication '%s' at URI '%s': '%s'",
		   pub->title, pub->path, problem);
      continue;
    }
    else {
      ap_log_error(APLOG_MARK, APLOG_INFO, 0, s,
		   "Publication '%s' at '%s' successfully loaded.",
		   pub->title, pub->path);
    }
  }

  /* a genxWriter used solely to look up character classes to support
   *  URI sanitization.  Never actually used to generate XML.
   */
  char_util = genxNew(NULL, NULL, NULL);

  return OK;
}

/* deal with AtomPub directive, which has to look like
 * AtomPub /base/of/uri/space /base/of/filesystem/space title author
 */
static const char * atompub_directive(cmd_parms * cmd, void * dummy,
				      int argc, char * const argv[])
{
  pub_t * pub;
  pub_t * publist;
  char * path;
  int i;

  if (argc != 4) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, cmd->server,
		 "AtompPub must have 4 arguments: URI path, directory, "
		 "title, and author");
    return NULL;
  }
  path = (char *) argv[0];

  /* O(N**2), sigh */
  publist = (pub_t *) atom_pubs->elts;
  for (i = 0; i < atom_pubs->nelts - 1; i++)
    if (is_prefix(path, publist[i].path) ||
	is_prefix(publist[i].path, path)) {
      ap_log_error(APLOG_MARK, APLOG_ERR, 0, cmd->server,
		   "One publication prefixes another: '%s' and '%s'",
		   path, publist[i].path);
      return NULL;
    }

  pub = (pub_t *) apr_array_push(atom_pubs);  
  pub->path = apr_pstrdup(cmd->pool, path);
  pub->dir = apr_pstrdup(cmd->pool, argv[1]);
  pub->title = apr_pstrdup(cmd->pool, argv[2]);
  pub->author = apr_pstrdup(cmd->pool, argv[3]);
  return NULL;
}

/***************** Request Time ***********************/

/* Translates r->uri into r->filename. 
 */
static int map_to_storage(request_rec * r)
{
  pub_t * pub = find_pub(r->uri);
  apr_finfo_t * finfo;
  apr_status_t status;

  if (pub == NULL)
    return DECLINED;

  r->filename =
    atom_join(r->pool, pub->dir, is_prefix(r->uri, pub->path), NULL);
  r->canonical_filename = r->filename;  /* why? */
    
  /* it seems we have to fill in r->finfo too */
  finfo = (apr_finfo_t *) apr_palloc(r->pool, sizeof(apr_finfo_t));
  status = apr_stat(finfo, r->filename,
		    APR_FINFO_MTIME|APR_FINFO_INODE|APR_FINFO_SIZE, r->pool);
  if (status != APR_SUCCESS) 
    return HTTP_NOT_FOUND;

  finfo->name = r->filename;
  r->finfo = *finfo;
  return OK;
}

/* Figure out the media type for our resources.  We toss some of the work back
 *  to httpd by applying .html and .xhtml extensions to some things as
 *  appropriate.
 */
static int type_checker(request_rec * r)
{
  pub_t * pub; 

  /* POST/PUT/DELETE, content-type depends on whether it worked */
  if (r->method_number != M_GET)
    return DECLINED;

  if ((pub = find_pub(r->uri)) == NULL)
    return DECLINED;

  /* It's a GET in our space; we have to do some work */

  /* if it's HTML/XHTML, let httpd use the extension */
  if (ap_regexec(pub->html_re, r->uri, 0, NULL, 0) == 0) {
    return DECLINED;
  }

  /* if it's an entry */
  if (ap_regexec(pub->entry_re, r->uri, 0, NULL, 0) == 0) {
    ap_set_content_type(r, "application/atom+xml;type=entry");
    return OK;
  }

  /* if it's a feed */
  if (ap_regexec(pub->feed_re, r->uri, 0, NULL, 0) == 0 ||
      ap_regexec(pub->collection_re, r->uri, 0, NULL, 0) == 0) {
    ap_set_content_type(r, "application/atom+xml");
    return OK;
  }
    
  /* if it's a service doc, APP media type */
  if (strcmp(pub->service_doc, r->uri) == 0) {
    ap_set_content_type(r, "application/atomsvc+xml");
    return OK;
  }

  /* If it's a media resource, append ".mt" to the end of the
   *  filename and look up the media-type, which we stashed in there.
   */
  if (ap_regexec(pub->media_re, r->uri, 0, NULL, 0) == 0) {
    char * mt_name = is_prefix(r->uri, pub->path);
    char * problem;
    char * media_type;

    mt_name = atom_join(r->pool, pub->dir, mt_name, NULL);
    mt_name = apr_psprintf(r->pool, "%s.%s", mt_name, MEDIA_TYPE_SUFFIX);
    if ((problem = atom_snarf(mt_name, &media_type, r->pool))) {
      ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		   "Can't snarf media-type file '%s': %s", mt_name,
		   problem);
      return HTTP_INTERNAL_SERVER_ERROR;
    }
    ap_set_content_type(r, media_type);
    return OK;
  }

  /* anything else is someone fishing around in our URI space.  Nope */
  return HTTP_NOT_FOUND;
}

static int handler(request_rec * r)
{
  pub_t * pub;


  /* GET should mostly be served out of a static file, whose
   *  media type should by now have been sorted out.  The exception is
   *  a collection with arguments, as in paging wanted.
   * PUT/POST/DELETE get special handling.
   * We don' need no steenking other methods.
   */
  if (r->method_number == M_GET && r->args == NULL)
    return DECLINED;

  if ((pub = find_pub(r->uri)) == NULL)
    return DECLINED;

  switch (r->method_number) {
  case M_GET: return get(r, pub);
  case M_POST: return post(r, pub);
  case M_PUT: return put(r, pub);
  case M_DELETE: return delete(r, pub);
  default: return DECLINED;
  }
}

/* only called if there are args on the URI */
static int get(request_rec * r, pub_t * pub) 
{
  int start = -1;
  int count = -1;
  char * a = r->args;
  char  * problem;
  genxElement * elements;
  genxAttribute * attributes;
  genxNamespace * namespaces;
  genxWriter gw;
  genxSender * sender;
  char * dirname;

  /* we only allow them on collection */
  if (ap_regexec(pub->collection_re, r->uri, 0, NULL, 0) != 0)
    return HTTP_NOT_FOUND;

  /* the only ones we allow are start= and count=
   * Shouldn't there be a utility to sort this crap out?
   */
  while (*a) {
    if (strncmp(a, "start=", 6) == 0) {
      a += 6;
      start = find_num(&a);
      if (start == -1)
	return HTTP_NOT_FOUND;
    }
    else if (strncmp(a, "count=", 6) == 0) {
      a += 6;
      count = find_num(&a);
      if (count == -1)
	return HTTP_NOT_FOUND;
    }
    else
      return HTTP_NOT_FOUND;

    if (*a == '&')
      a++;
  }

  if (count == -1)
    count = DEFAULT_ENTRIES_IN_FEED;
  if (start == -1)
    start = 0;

  /* OK, fire the collection down the pipe */
  problem = atom_genx_client_writer(r->pool, r, &gw, &sender,
				    &elements, &attributes, &namespaces);
  if (problem) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't create XML writer");
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  dirname = apr_pstrdup(r->pool, r->uri);
  dirname = atom_join(r->pool, pub->dir, is_prefix(r->uri, pub->path), NULL);
  *(rindex(dirname, '/') + 1) = 0;
  problem = generate_collection(dirname, r->pool, r->server,  pub, gw, sender,
				start, count, elements, attributes, namespaces);
  if (problem) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't generate collection");
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  return OK;
}

static int post(request_rec * r, pub_t * pub) 
{
  char * media_type;
  entry_t entry;

  entry.pub = pub;

  /* has to be a collection */
  if (ap_regexec(entry.pub->collection_re, r->uri, 0, NULL, 0) != 0) {
    /* grr... required by 2616.
     * How should I know which methods are allowed on their bogus URI
     */
    apr_table_set(r->headers_out, "Allow", "GET, HEAD");
    return HTTP_METHOD_NOT_ALLOWED;
  }

  /* OK, this is an attempt to post to a collection.
   * - Check that the content type is OK for this collection
   */
  media_type = (char *) apr_table_get(r->headers_in, "Content-Type");
  if (media_type == NULL)
    return HTTP_UNSUPPORTED_MEDIA_TYPE;
  if (is_prefix(r->uri, entry.pub->entries_prefix)) {
    if (strcmp(media_type, "application/atom+xml;type=entry") &&
	strcmp(media_type, "application/atom+xml"))
      return HTTP_UNSUPPORTED_MEDIA_TYPE;
    return post_entry(r, &entry);
  }
  else if (is_prefix(r->uri, entry.pub->media_prefix)) {
    return post_media(r, entry.pub);
  }
  else {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Weird media type '%s' on POST '%s'", media_type, r->uri);
    return HTTP_INTERNAL_SERVER_ERROR; /* shouldn't happen */
  }
}

/* Update an entry.  We have to do a trick here to enable producing collections
 *  by walking the file tree.  When we PUT to an entry in yyyy/mm/dd, we create
 *  a file named L.<random-numbers> in today's yyyy/mm/dd which contains only
 *  a link to the real name of the file we just updated with PUT.  The
 *  tree-walker in iterator.c knows about these things.  Of course, they can
 *  get out of date if there are subsequent PUTs or DELETEs, but that's OK,
 *  the iterator notices this and cleans up as it goes.
 */
static int put(request_rec * r, pub_t * pub) 
{
  apr_global_mutex_t * mutex;
  apr_status_t status;
  char * media_type;
  char * etag;
  char * if_match;
  int is_entry;
  apr_finfo_t finfo;
  entry_t entry;
  char * problem;
  
  /* may not be necessary */
  if (!atom_file_exists(r->filename, r->pool))
    return HTTP_NOT_FOUND;

  /* gotta have that conditional put */
  if_match = (char *) apr_table_get(r->headers_in, "If-Match");
  if (if_match == NULL || if_match[0] == '*') {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Attempt to PUT to '%s' without an If-Match header",
		 r->uri);
    apr_table_set(r->headers_out, "Allow", "GET, HEAD");
    return HTTP_METHOD_NOT_ALLOWED;
  }

  /* entry? */
  if (ap_regexec(pub->entry_re, r->uri, 0, NULL, 0) == 0) {

    /* XXX TODO
     * For the moment, no PUT to an MLE.  Down the road, we should allow
     *  them to provide summary/content/cat/title/author.  But for the
     *  the time being, we'll autogenerate the MLE from the media POSTs and
     *  PUTs.
     */
    if (is_prefix(r->uri, pub->media_prefix)) {
      apr_table_set(r->headers_out, "Allow", "GET, HEAD, DELETE");
      return HTTP_METHOD_NOT_ALLOWED;
    }

    /* sanity-check media-type */
    media_type = (char *) apr_table_get(r->headers_in, "Content-Type");
    if (media_type == NULL)
      return HTTP_UNSUPPORTED_MEDIA_TYPE;

    if (strcmp(media_type, "application/atom+xml;type=entry") &&
	strcmp(media_type, "application/atom+xml"))
      return HTTP_UNSUPPORTED_MEDIA_TYPE;

    is_entry = 1;

    /* parse the incoming PUT body */
    if ((status = ap_xml_parse_input(r, &(entry.xml))) != OK) {
      ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		   "Can't parse XML PUT body");
      return HTTP_INTERNAL_SERVER_ERROR;
    }
  }
  else if (ap_regexec(pub->media_re, r->uri, 0, NULL, 0) == 0) {
    is_entry = 0;
  }
  else {
    apr_table_set(r->headers_out, "Allow", "GET, HEAD");
    return HTTP_METHOD_NOT_ALLOWED;
  }

  /* Sigh.  We can create entries without locking, I'm pretty sure, but
   *  I think you can only have one thread/process at a time looking at
   *  the ETag and doing the PUT.
   */
  status = apr_global_mutex_create(&mutex, r->filename, APR_LOCK_DEFAULT,
				   r->pool);
  if (status != APR_SUCCESS) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't create mutex for '%s' to do PUT",
		 r->filename);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  /* I would rather use global_mutex_trylock and simply return if I can't
   *  get the lock, but that turns out not to be supported on certain
   *  platforms.
   */
  status = apr_global_mutex_lock(mutex);
  if (status != APR_SUCCESS) {

    /* XXX this hack is a reminder of how we're going to capture
     *  human-readable strings when we put in dav-style error handling
     */
    char buf[1000];
    char * msg = apr_strerror(status, buf, 1000);
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't lock '%s' for PUT: '%s'", r->filename, msg);
    apr_global_mutex_destroy(mutex);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  /* ETags */
  /* It turns out that make_etag depends on having the inode, size, and
   *  mtime filled in, which in the normal processing cycle, they aren't
   *  at this point.
   */
  /* Apache generates weak etags during the first second of a file's life,
   *  and if the client fetched immediately after posting, may have one
   *  of those.  Based on the the language in 2616, a weak ETag match
   *  is good enough for us, so we de-weakify both the If-Match value
   *  and the file ETag.
   */
  status = apr_stat(&finfo, r->filename,
		    APR_FINFO_MTIME|APR_FINFO_INODE|APR_FINFO_SIZE, r->pool);
  r->mtime = r->finfo.mtime = finfo.mtime;
  r->finfo.size = finfo.size;
  r->finfo.inode = finfo.inode;
  etag = ap_make_etag(r, 0);
  if (strncmp(if_match, WEAK_PREFIX, strlen(WEAK_PREFIX)) == 0)
    if_match += strlen(WEAK_PREFIX);
  if (strncmp(etag, WEAK_PREFIX, strlen(WEAK_PREFIX)) == 0)
    etag += strlen(WEAK_PREFIX);
  if (strcmp(if_match, etag)) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "ETag mismatch on PUT (%s vs %s), rejected", if_match, etag);
    apr_global_mutex_unlock(mutex);
    apr_global_mutex_destroy(mutex);
    return HTTP_PRECONDITION_FAILED;
  }

  if (is_entry) 
    status = put_entry(r, pub, &entry);
  else
    status = put_media(r, pub, &entry);

  apr_global_mutex_unlock(mutex);
  apr_global_mutex_destroy(mutex);

  /* put a timestamp file in place */
  if (status == OK) {
    char * linkname;
    linkname = atom_join(r->pool, entry.coll_base,
			 yyyymmdd_at(apr_time_now(), r->pool), NULL);
    linkname = apr_psprintf(r->pool, "%s/%s%ld-%ld",
			    linkname, LINK_PREFIX, random(), random());
    if ((problem = to_file(linkname, entry.filename, r->pool))) {
      ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		   "Can't create link file '%s' to store '%s': %s", linkname,
		   entry.filename, problem);
      return HTTP_INTERNAL_SERVER_ERROR;
    }
  }

  /* rewrite the collection */
  problem = write_collection(r->pool, r->server, pub, entry.coll_base,
			     DEFAULT_ENTRIES_IN_FEED);
  if (problem) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't write collection in '%s' post-PUT: %s",
		 entry.coll_base, problem);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  return status;
}
static int delete(request_rec * r, pub_t * pub) 
{
  apr_status_t status;
  char * problem;
  char * coll_base;

  /* may not be necessary */
  if (!atom_file_exists(r->filename, r->pool))
    return HTTP_NOT_FOUND;

  /* has to be an entry */
  if (ap_regexec(pub->entry_re, r->uri, 0, NULL, 0) != 0) {
    apr_table_set(r->headers_out, "Allow", "GET, HEAD");
    return HTTP_METHOD_NOT_ALLOWED;
  }

  /* can't delete media resources directly */
  if (ap_regexec(pub->media_re, r->uri, 0, NULL, 0) == 0) {
    apr_table_set(r->headers_out, "Allow", "GET, HEAD, PUT");
    return HTTP_METHOD_NOT_ALLOWED;
  }

  /* if it's a media-link entry, then find & remove the associated
   *  media resource.
   */
  if (is_prefix(r->uri, pub->media_prefix)) {
    apr_file_t * fp;
    apr_xml_parser * parser;
    apr_xml_doc * mle;
    apr_xml_elem * child;
    apr_xml_attr * attr;
    char * * namespaces;

    status = apr_file_open(&fp, r->filename, APR_FOPEN_READ, PERMS, r->pool);
    if (status != APR_SUCCESS) {
      ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		   "Can't open media link entry in '%s'", r->filename);
      return HTTP_INTERNAL_SERVER_ERROR;
    }
    status= apr_xml_parse_file(r->pool, &parser, &mle, fp, BUFSIZ);
    if (status != APR_SUCCESS) {
      ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		   "Can't parse media link entry in '%s'", r->filename);
      return HTTP_INTERNAL_SERVER_ERROR;
    }
    namespaces = (char * *) mle->namespaces->elts;
    for (child = mle->root->first_child; child; child = child->next) {
      if (child->ns != APR_XML_NS_NONE &&
	  strcmp(namespaces[child->ns], ATOM_NS) == 0 &&
	  strcmp(child->name, "content") == 0) {
	for (attr = child->attr; attr; attr = attr->next) {
	  if (attr->ns == APR_XML_NS_NONE &&
	      strcmp(attr->name, "src") == 0) {
	    char * src = (char *) attr->value;
	    src =
	      atom_join(r->pool, pub->dir, is_prefix(src, pub->path), NULL);
	    if (apr_file_remove(src, r->pool) != APR_SUCCESS) {
	      ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
			   "Can't remove media resource at '%s'", src);
	      return HTTP_INTERNAL_SERVER_ERROR;
	    }
	    src = apr_psprintf(r->pool, "%s.%s", src, MEDIA_TYPE_SUFFIX);
	    if (apr_file_remove(src, r->pool) != APR_SUCCESS) {
	      ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
			   "Can't remove media resource at '%s'", src);
	      return HTTP_INTERNAL_SERVER_ERROR;
	    }
	    break;
	  }
	}
	break;
      }
    }
  }

  /* this seems to be altogether too easy */
  if ((status = apr_file_remove(r->filename, r->pool)) != APR_SUCCESS) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't remove file '%s'", r->filename);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  coll_base = find_coll_base(r, pub);
  if ((problem = write_collection(r->pool, r->server, pub, coll_base,
				  DEFAULT_ENTRIES_IN_FEED)) != NULL) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't write collection in '%s': %s", coll_base, problem);
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  
  return OK;
}

/********************* Utilities ***********************/

static int find_num(char * * a_p) 
{
  char * a = *a_p;

  int sofar = 0;
  while (isdigit(*a)) {
    sofar = (sofar * 10) + (*a - '0');
    a++;
  }
  if (*a == 0 || *a == '&') {
    *a_p = a;
    return sofar;
  }
  else
    return -1;
}

/* Have to be careful that this gets run once-only for each collection/entry
 */
static char * new_id(apr_pool_t * pool) 
{
  apr_uuid_t uuid;
  char buf[APR_UUID_FORMATTED_LENGTH + 1];
  apr_uuid_get(&uuid);
  apr_uuid_format(buf, &uuid);
  return apr_psprintf(pool, "urn:uuid:%s", buf);
}

static char * find_coll_base(request_rec * r, pub_t * pub) 
{
  char * coll;
  int i;
  int slashes;

  /* stupid collection.
   * we have to bypass the pub prefix, then three steps: ATOM_DIR,
   *  ENTRIES_DIR, and the collection name;
   */
  coll = apr_pstrdup(r->pool, is_prefix(r->uri, pub->path));
  for (i = slashes = 0; slashes < 4; i++)
    if (coll[i] == '/')
      slashes++;
  coll[i] = 0;
  return atom_join(r->pool, pub->dir, coll, NULL);
}

/* Ouch, 11 arguments.  Major refactoring signal, n'est-ce pas?
 */
static char * generate_collection(char * dirname, apr_pool_t * pool,
				  server_rec * s,
				  pub_t * pub,
				  genxWriter gw,
				  genxSender * sender,
				  int start,
				  int count,
				  genxElement * elements,
				  genxAttribute * attributes,
				  genxNamespace * namespaces)
{
  char * id_name = atom_join(pool, dirname, ID_FILE, NULL);
  int entries_in_feed;
  char * entry_name;
  char * * entry_names;
  int i;
  void * iter;
  char * problem;
  char * prefix;
  char * id;

  /* Unique ID for the collection */
  if ((problem = atom_snarf(id_name, &id, pool)))
    return apr_psprintf(pool, "Can't read ID file '%s'", id_name);

  /* write the per-feed elements */
  genxStartElement(elements[ATOM_FEED_E]);
  atom_genx_simple_element(gw, elements[ATOM_TITLE_E], pub->title);

  prefix = is_prefix(dirname, pub->dir) + 1 + strlen(ATOM_DIR);
  genxStartElement(elements[ATOM_LINK_E]);
  genxAddAttribute(attributes[HREF_A],
		   (utf8) atom_join(pool, pub->path, PUB_DIR, prefix, NULL));
  genxEndElement(gw);
  atom_genx_simple_element(gw, elements[ATOM_UPDATED_E],
			   atom_time_at(apr_time_now(), pool));
  genxStartElement(elements[ATOM_AUTHOR_E]);
  atom_genx_simple_element(gw, elements[ATOM_NAME_E], pub->author);
  genxEndElement(gw);
  atom_genx_simple_element(gw, elements[ATOM_ID_E], id);

  if ((problem = atom_iter_new(pool, dirname, &iter)))
    return apr_psprintf(pool, "Can't start iterator on '%s': %s", dirname,
			problem);

  /* we have to save up the entry filenames first, because we don't know
   *  whether to write a <link rel="next" into the feed header until we know
   *  if we're going to get all the entries in the feed.
   */
  entry_names = (char * *) apr_pcalloc(pool, count * sizeof(char *));

  for (problem = atom_iter_next(iter, &entry_name), i = entries_in_feed = 0;
       entry_name && (!problem) && (entries_in_feed < (start + count));
       problem = atom_iter_next(iter, &entry_name), entries_in_feed++) {
    if (entries_in_feed >= start)
      entry_names[i++] = entry_name;
  }

  if (entries_in_feed == (start + count)) {
    char * uri = atom_join(pool, pub->path, is_prefix(dirname, pub->dir),
			   COLLECTION_FILE, NULL);
    genxStartElement(elements[ATOM_LINK_E]);
    genxAddAttribute(attributes[REL_A], (utf8) "next");
    genxAddAttribute(attributes[HREF_A],
		     (utf8) apr_psprintf(pool, "%s?start=%d", uri,
					 start + count));
    genxEndElement(gw);
  }

  /* Now we jam in the entries, by snarfing them one at a time
   *  in reverse chronological order and just jamming them into the
   *  feed as-is.
   */
  for (count = i, i = 0; i < count; i++)
  {
    char * entry_body;

    if ((problem = atom_snarf(entry_names[i], &entry_body, pool))) 
      return apr_psprintf(pool, "Can't read entry '%s': %s", entry_names[i],
			  problem);

    if (sender->send(genxGetUserData(gw), (utf8) entry_body) != GENX_SUCCESS)
      return "Can't write to collection";

  }
  if (problem)
    return apr_psprintf(pool, "Error iterating collection '%s': %s", dirname,
			problem);

  genxEndElement(gw);
  genxEndDocument(gw);
  return NULL;
}


/* The dirname is where to write it, the entries in the collection will
 *  be in yyyy/mm/dd directories underneath.
 * There's not much concern for concurrency; we write the collection into
 *  another file and rely on apr_file_rename being close enough to the
 *  filesystem to let anyone who needs to update it just go ahead and do
 *  so.
 * It might be smarter to just set a flag saying "collection needs re-writing"
 *  and let that happen on the next GET.
 */
static char * write_collection(apr_pool_t * pool, server_rec * s,
			       pub_t * pub, char * dirname, int size)
{
  char * filename = atom_join(pool, dirname, COLLECTION_FILE, NULL);
  char * tempname = apr_psprintf(pool, "%s%ld-%ld",
				 atom_join(pool, dirname, TEMP_PREFIX),
				 random(), random());
  apr_file_t * fp;
  char * problem;
  apr_status_t status;
  genxElement * elements;
  genxAttribute * attributes;
  genxNamespace * namespaces;
  genxWriter gw;
  genxSender * sender;

#if defined(UNIT_TEST)
  /* pardon the ugliness, have to test this under controlled conditions */
  if (suppress_writing_collections())
    return NULL;
#endif

  status = apr_file_open(&fp, tempname, APR_FOPEN_WRITE|APR_FOPEN_CREATE,
			 PERMS, pool);
  if (status != APR_SUCCESS) {
    return apr_psprintf(pool, "Can't open collection tempfile '%s'", tempname);
  }

  problem = atom_genx_file_writer(pool, fp, &gw, &sender,
				  &elements, &attributes, &namespaces);
  if (problem) {
    return apr_psprintf(pool, "Can't start XML writer: '%s'", problem);
  }

  problem = generate_collection(dirname, pool, s, pub, gw, sender,
				0, size, elements, attributes, namespaces);
  apr_file_close(fp);
  if (problem)
    return apr_psprintf(pool, "Can't generate collection in '%s'", dirname);
  
  if (apr_file_rename(tempname, filename, pool) != APR_SUCCESS)
    return apr_psprintf(pool, "Can't rename '%s' to '%s'", tempname, filename);
  return NULL;
}

static int put_media(request_rec * r, pub_t * pub, entry_t * entry)
{
  media_t media;
  int status;
  ap_regmatch_t regm[3];
  char * slash;

  media.mle = entry;
  media.mle->pub = pub;
  media.mle->coll_base = find_coll_base(r, pub);

  /* compute directories and names */
  if (ap_regexec(pub->mle_maker, r->uri, 3, regm, 0)) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "PUT media URI '%s' doesn't match mle-maker regex", r->uri);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  media.mle_path = ap_pregsub(r->pool, "$1$2", r->uri, 3, regm);
  media.mle_path = is_prefix(media.mle_path, pub->path);
  slash = rindex(media.mle_path, '/');
  media.filename = apr_pstrdup(r->pool, slash + 1);
  *slash = 0;
  media.bits_path = apr_pstrdup(r->pool, is_prefix(r->uri, pub->path));
  *(rindex(media.bits_path, '/')) = 0;

  if ((status = replace_media(r, &media)) != OK)
    return status;

  /* build and save a Media Link Entry for the bits */
  if ((status = replace_mle(r, &media)) != OK)
    return status;

  return OK;
}

/* build an empty DOM with all the namespaces set up and so on */
static apr_xml_doc * make_mle_shell(apr_pool_t * pool) 
{
  char * * ns_here;
  apr_xml_doc * doc;
  int atom_ns = 1;

  doc = (apr_xml_doc *) apr_pcalloc(pool, sizeof(apr_xml_doc));
  doc->namespaces = apr_array_make(pool, 3, sizeof(char *));

  /* first slot hard-wired for DAV */
  apr_array_push(doc->namespaces);
  ns_here = (char * *) apr_array_push(doc->namespaces);
  *ns_here = ATOM_NS;
  ns_here = (char * *) apr_array_push(doc->namespaces);
  *ns_here = APP_NS;
  ns_here = (char * *) apr_array_push(doc->namespaces);
  *ns_here = XHTML_NS;

  doc->root = (apr_xml_elem *) apr_pcalloc(pool, sizeof(apr_xml_elem));
  doc->root->name = "entry";
  doc->root->ns = atom_ns;
  return doc;
}

static int save_mle(request_rec * r, media_t * media)
{
  char * mtime = atom_time_at(r->request_time, r->pool);
  apr_xml_elem * root;
  int atom_ns = 1;
  int app_ns = 2;
  int xhtml_ns = 3;
  char * media_type;
  char * title;
  apr_xml_elem * child;
  char * tempname;
  char * target;
  char * problem;
  pub_t * pub = media->mle->pub;
  char * bits_uri =
    atom_join(r->pool, pub->path, media->bits_path, media->filename, NULL);
  char * mle_uri = 
    atom_join(r->pool, pub->path, media->mle_path, media->filename, NULL);

  media->mle->xml = make_mle_shell(r->pool);
  root = media->mle->xml->root;

  insert_simple_child(r, root, atom_ns, "updated", mtime);
  insert_simple_child(r, root, app_ns, "edited", mtime);
  insert_simple_child(r, root, atom_ns, "id", media->mle->id);
  media_type = (char *) apr_table_get(r->headers_in, "Content-Type");
  title = apr_psprintf(r->pool, "%s: %d bytes at %s",
		       media_type, media->size, mtime);
  insert_simple_child(r, root, atom_ns, "title", title);
  child = insert_simple_child(r, root, atom_ns, "author", NULL);
  insert_simple_child(r, child, atom_ns, "name", pub->author);

  /* don't need an rel="alt' link */
  child = insert_simple_child(r, root, atom_ns, "link", NULL);
  insert_attr(r, child, "rel", "edit");
  insert_attr(r, child, "href", mle_uri);
  child = insert_simple_child(r, root, atom_ns, "link", NULL);
  insert_attr(r, child, "rel", "edit-media");
  insert_attr(r, child, "href", bits_uri);

  /* construct a nice summary */
  child = insert_simple_child(r, root, atom_ns, "summary", NULL);
  insert_attr(r, child, "type", "xhtml");
  child = insert_simple_child(r, child, xhtml_ns, "div", NULL);
  insert_simple_child(r, child, xhtml_ns, "p",
		      apr_psprintf(r->pool, "Media-type: '%s'", media_type));
  insert_simple_child(r, child, xhtml_ns, "p",
		      apr_psprintf(r->pool, "Size: '%d'", media->size));
  insert_simple_child(r, child, xhtml_ns, "p",
		      apr_psprintf(r->pool, "Posted-at: '%s'", mtime));

  /* content and we're done */
  child = insert_simple_child(r, root, atom_ns, "content", NULL);
  insert_attr(r, child, "type", media_type);
  insert_attr(r, child, "src", bits_uri);

  /* blast it into a tempfile */
  tempname = atom_join(r->pool, pub->dir, media->mle_path, NULL);
  tempname = apr_psprintf(r->pool, "%s/%s%ld-%ld", tempname, 
			  TEMP_PREFIX, random(), random());
  if ((problem = write_entry(r, media->mle->xml, tempname))) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't write Media Link Entry into '%s': %s", tempname,
		 problem);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  /* move it to its final home */
  target =
    atom_join(r->pool, pub->dir, media->mle_path, media->filename, NULL);
  if (apr_file_rename(tempname, target, r->pool) != APR_SUCCESS) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't rename from '%s' to '%s'", tempname, target);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  if (r->parsed_uri.port_str != NULL)
    media->mle->uri = apr_psprintf(r->pool, "http://%s:%s%s/%s/%s",
				   r->hostname, r->parsed_uri.port_str,
				   pub->path, media->mle_path, media->filename);
  else
    media->mle->uri = apr_psprintf(r->pool, "http://%s%s/%s/%s",
				   r->hostname, 
				   pub->path, media->mle_path, media->filename);

  return OK;
}

/* mle stands for Media Link Entry.  We are updating a media object and
 *  replacing its corresponding MLE.  One field, the ID, has to be copied
 *  in from the existing MLE.
 */
static int replace_mle(request_rec * r, media_t * media)
{
  char * * namespaces;
  apr_xml_elem * child;
  apr_file_t * fp;
  apr_status_t status;
  char * filename;
  apr_xml_parser * parser;
  apr_xml_doc * existing;
  char * id = "";

  /* get the ID from the old MLE */
  filename = atom_join(r->pool, media->mle->pub->dir,
		       media->mle_path, media->filename, NULL);
  status = apr_file_open(&fp, filename, APR_FOPEN_READ, PERMS, r->pool);
  if (status != APR_SUCCESS) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't open existing MLE file '%s'", filename);
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  media->mle->filename = filename;

  status = apr_xml_parse_file(r->pool, &parser, &existing, fp, BUFSIZ);
  if (status != APR_SUCCESS) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't parse existing MLE file '%s'", filename);
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  apr_file_close(fp);

  namespaces = (char * *) existing->namespaces->elts;
  for (child = existing->root->first_child; child; child = child->next) {
    /* in the atom namespace */
    if (strcmp(namespaces[child->ns], ATOM_NS) == 0) {
      if (strcmp(child->name, "id") == 0) {
	apr_text * text;
	
	for (text = child->first_cdata.first; text; text = text->next)
	  id = apr_pstrcat(r->pool, id, text->text, NULL);
	break;
      }
    }
  }

  media->mle->id = id;
  return save_mle(r, media);
}

/* mle stands for Media Link Entry.  The easiest way to do this is to build it
 *  in XML so we can just pass it to write_entry to persist it.
 */
static int save_new_mle(request_rec * r, media_t * media) 
{
  media->mle->id = new_id(r->pool);
  return save_mle(r, media);
}


/* Posting a media object.  We have to
 * - find a place for the bits
 * - remember the media type so we can look it up later
 * - create a synthetic atom media link entry in the media collection
 * - update the collection feed for the media collection
 */
static int post_media(request_rec * r, pub_t * pub) 
{
  media_t media;
  entry_t mle;
  int status;
  char * mle_filename;
  char * mle_body;
  char * problem;

  /* save the bits */
  media.mle = &mle;
  mle.pub = pub;
  if ((status = save_new_media(r, &media)) != OK)
    return status;

  /* build and save a Media Link Entry for the bits */
  if ((status = save_new_mle(r, &media)) != OK)
    return status;

  /* read back the mle */
  mle_filename = atom_join(r->pool, pub->dir, media.mle_path,
			   media.filename, NULL);
  if ((problem = atom_snarf(mle_filename, &mle_body, r->pool))) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't snarf '%s': %s", mle_filename, problem);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  /* set up the headers */
  apr_table_set(r->headers_out, "Location", mle.uri);
  apr_table_set(r->headers_out, "Content-Location", mle.uri);
  ap_set_content_type(r, "application/atom+xml;type=entry");
  r->status = 201;

  /* stuff the MLE back down the pipe to the client */
  ap_rputs(mle_body, r);

  /* re-write the collection */
  media.mle->coll_base = find_coll_base(r, pub);
  if ((problem = write_collection(r->pool, r->server, pub, media.mle->coll_base,
				  DEFAULT_ENTRIES_IN_FEED))) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't write media collection into '%s': %s",
		 media.mle->coll_base, problem);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  return OK;
}

/* update an entry in place
 * It's remarkably similar to initial entry creation, except for we
 *  don't change any of the URIs, and we re-use the unique ID from
 *  the entry that's already there.  So you could PUT a version with
 *  a completely new title and no similarity with the existing content.
 */
static int put_entry(request_rec * r, pub_t * pub, entry_t * entry)
{
  apr_status_t status;
  apr_file_t * fp;
  apr_xml_parser * parser;
  apr_xml_doc * existing;
  apr_xml_elem * child;
  char * * namespaces;
  char * problem;
  char * tempname;

  entry->html_name = NULL;

  status = apr_file_open(&fp, r->filename, APR_FOPEN_READ, PERMS, r->pool);
  if (status != APR_SUCCESS) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't open entry file '%s' for reading", r->filename);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  status = apr_xml_parse_file(r->pool, &parser, &existing, fp, BUFSIZ);
  if (status != APR_SUCCESS) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't parse entry file '%s'", r->filename);
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  apr_file_close(fp);

  /* pull the ID & HTML name out of the existing child */
  namespaces = (char * *) existing->namespaces->elts;
  for (child = existing->root->first_child; child; child = child->next) {
    /* in the atom namespace */
    if (strcmp(namespaces[child->ns], ATOM_NS) == 0) {
      if (strcmp(child->name, "id") == 0) {
	apr_text * text;
	
	entry->id = "";
	for (text = child->first_cdata.first; text; text = text->next)
	  entry->id = apr_pstrcat(r->pool, entry->id, text->text, NULL);
      }
      else if (entry->html_name == NULL &&
	       strcmp(child->name, "link") == 0) {
	apr_xml_attr * attr;
	char * href = NULL;
	char * rel = NULL;
	for (attr = child->attr; attr; attr = attr->next) {
	  if (attr->ns == APR_XML_NS_NONE) {
	    if (strcmp(attr->name, "href") == 0)
	      href = (char *) attr->value;
	    else if (strcmp(attr->name, "rel") == 0)
	      rel = (char *) attr->value;
	  }
	}
	if (rel == NULL || strcmp(rel, "alt") == 0)
	  entry->html_name = href;
      }
    }
  }

  entry->filename = r->filename;
  entry->uri = r->uri;
  entry->pub = pub;

  entry->coll_base = find_coll_base(r, pub);

  /* Fix up as required (ID, dates, links, etc) */
  if ((problem = prepare_entry(r, entry)) != NULL) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Problem preparing entry via PUT: %s", problem);
    return HTTP_FORBIDDEN;
  }

  /* write it to a temp file */
  tempname = apr_pstrdup(r->pool, entry->filename);
  *(rindex(tempname, '/')) = 0;
  tempname = apr_psprintf(r->pool, "%s/%s%ld-%ld", tempname,
			  TEMP_PREFIX, random(), random());
  if ((problem = write_entry(r, entry->xml, tempname)) != NULL) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't write PUT entry into '%s': %s", tempname, problem);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  /* put it where it belongs */
  if (apr_file_rename(tempname, entry->filename, r->pool) != APR_SUCCESS) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't rename from '%s' to '%s'", tempname, entry->filename);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  /* there's more work to do, but this is all we need to do while
   *  the entry's locked.
   */
  return OK;
}

/* Create an entry, and generate related URIs */
static int post_entry(request_rec * r, entry_t * entry) 
{
  int status;
  char * problem;
  char * entry_body = NULL;

  /* Try to suck it up */
  if ((status = ap_xml_parse_input(r, &(entry->xml))) != OK) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't parse XML POST body");
    return status;
  }

  /* Blast it into a file */
  status = save_new_entry(r, entry);
  if (status != OK)
    return status;

  /* read the file back */
  if ((problem = atom_snarf(entry->filename, &entry_body, r->pool))) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't snarf '%s': %s", entry->filename, problem);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  /* headers */
  apr_table_set(r->headers_out, "Location", entry->uri);
  apr_table_set(r->headers_out, "Content-Location", entry->uri);
  ap_set_content_type(r, "application/atom+xml;type=entry");
  r->status = 201;

  /* blast it down the pipe */
  ap_rputs(entry_body, r);

  /* update the collection */
  if ((problem = write_collection(r->pool, r->server, entry->pub,
				  entry->coll_base,
				  DEFAULT_ENTRIES_IN_FEED))) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't write collection in '%s': %s",
		 entry->coll_base, problem);
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  return OK;
}

/* Generate an atom-conformant timestamp */
static char * atom_time_at(apr_time_t when, apr_pool_t * pool) 
{
  char * buf = apr_pcalloc(pool, 100);
  apr_size_t time_len;
  apr_time_exp_t times;

  apr_time_exp_lt(&times, when);
  apr_strftime(buf, &time_len, 100, "%Y-%m-%dT%H:%M:%S%z", &times);
  
  /* I'm assuming there's a strftime incantation to do this? */
  buf[time_len] = buf[time_len - 1];
  buf[time_len - 1] = buf[time_len - 2];
  buf[time_len - 2] = ':';

  return buf;
}

/* Drop in a child element with no attributes and possibly content */
static apr_xml_elem * insert_simple_child(request_rec * r,
					  apr_xml_elem * parent,
					  int ns_ind, char * type,
					  char * content) 
{
  apr_xml_elem * e =
    (apr_xml_elem *) apr_pcalloc(r->pool, sizeof(apr_xml_elem));

  e->ns = ns_ind;
  e->name = type;
  e->parent = parent;
  e->next = parent->first_child;
  e->following_cdata = parent->first_cdata;
  if (content) 
    insert_text_content(r, e, content);
  else 
    parent->first_cdata.first = parent->first_cdata.last = NULL;
  parent->first_child = e;
  return e;
}

/* Drop #PCData into an element */
static char * insert_text_content(request_rec * r, apr_xml_elem * e, char * t) 
{
  apr_text_header * text_h = apr_palloc(r->pool, sizeof(apr_text_header));
  apr_text * text = apr_palloc(r->pool, sizeof(apr_text));

  text->text = t;
  text->next = NULL;
  text_h->first = text_h->last = text;
  e->first_cdata = *text_h;
  return NULL;
}

/* Add an attribute to an element */
static void insert_attr(request_rec * r, apr_xml_elem * e, 
			char * name, char * value) 
{
  apr_xml_attr * attr;

  for (attr = e->attr; attr; attr = attr->next)
    if (strcmp(name, attr->name) == 0)
      break;

  if (attr == NULL) {
    attr = apr_pcalloc(r->pool, sizeof(apr_xml_attr)); 
    attr->ns = APR_XML_NS_NONE;
    attr->name = name;
    attr->next = e->attr;
    e->attr = attr;
  }
  attr->value = value;
}

/* Entry validation/fix-up; return NULL or the decription of the problem
 * - must have a title
 * - add an author if we don't know it
 * - fix up atom:updated, atom:id, app:edited
 * - has to be a content, with sane type= attribute
 */
static char * prepare_entry(request_rec * r, entry_t * entry)
{
  apr_xml_elem * child;
  char * * ns_names = (char * *) entry->xml->namespaces->elts;
  int i;
  int atom_ns = -1, app_ns = -1;
  apr_xml_elem * updated = NULL, * author = NULL, * content = NULL,
    * id = NULL, * edited = NULL, * alt_link = NULL, * edit_link = NULL;
  apr_xml_elem * root = entry->xml->root;
  char * mtime;

  for (i = 0; i < entry->xml->namespaces->nelts; i++) {
    if (!strcmp(ns_names[i], ATOM_NS))
      atom_ns = i;
    else if (!strcmp(ns_names[i], APP_NS))
      app_ns = i;
  }
  /* probably no app stuff in the incoming */
  if (app_ns == -1) {
    char * * ns_here;
    app_ns = entry->xml->namespaces->nelts;
    ns_here = (char * *) apr_array_push(entry->xml->namespaces);
    *ns_here = APP_NS;
    ns_names = (char * *) entry->xml->namespaces->elts;
  }

  if (root->ns != atom_ns || strcmp(root->name, "entry") != 0) 
    return "Root of doc posted as an Entry was not atom:entry";

  for (child = root->first_child; child; child = child->next) {
    if (child->ns == atom_ns) {
      if (strcmp(child->name, "author") == 0)
	author= child;
      else if (strcmp(child->name, "content") == 0)
	content = child;
      else if (strcmp(child->name, "updated") == 0)
	updated = child;
      else if (strcmp(child->name, "id") == 0)
	id = child;
      else if (strcmp(child->name, "link") == 0) {
	apr_xml_attr * attr;
	apr_xml_attr * rel = NULL;
	for (attr = child->attr; attr; attr = attr->next) {
	  if (strcmp(attr->name, "rel") == 0)
	    rel = attr;
	}
	if (rel == NULL)
	  alt_link = child;
	else if (strcmp(rel->value, "alt") == 0)
	  alt_link = child;
	else if (strcmp(rel->value, "edit") == 0)
	  edit_link = child;
      }
    }
    else if (child->ns == app_ns && strcmp(child->name, "edited") == 0)
      edited = child;
  }

  mtime = atom_time_at(r->request_time, r->pool);
  if (updated == NULL) 
    updated = insert_simple_child(r, root, atom_ns, "updated", NULL);
  insert_text_content(r, updated, mtime);
  if (edited == NULL)
    edited = insert_simple_child(r, root, app_ns, "edited", NULL);
  insert_text_content(r, edited, mtime);

  if (id == NULL)
    id = insert_simple_child(r, root, atom_ns, "id", NULL);
  insert_text_content(r, id, entry->id);

  if (author == NULL) {
    author = insert_simple_child(r, root, atom_ns, "author", NULL);
    author = insert_simple_child(r, author, atom_ns, "name", NULL);
    insert_text_content(r, author, entry->pub->author);
  }

  if (alt_link == NULL)
    alt_link = insert_simple_child(r, root, atom_ns, "link", NULL);
  insert_attr(r, alt_link, "href", entry->html_name);

  if (edit_link == NULL) {
    edit_link = insert_simple_child(r, root, atom_ns, "link", NULL);
    insert_attr(r, edit_link, "rel", "edit");
  }
  insert_attr(r, edit_link, "href", entry->uri);

  return NULL;
}

/* write out an element of a document, then its children recursively */
static char * write_element(apr_xml_doc * doc, apr_xml_elem * e, genxWriter gw)
{
  apr_xml_attr * attr;
  utf8 * ns_names = (utf8 *) doc->namespaces->elts;
  apr_xml_elem * child;
  apr_text * text;

  genxStartElementLiteral(gw, ns_names[e->ns], (utf8) e->name);
  if (e->lang)
    genxAddAttributeLiteral(gw, (utf8) XML_NS, (utf8) "lang", (utf8) e->lang);

  for (attr = e->attr; attr; attr = attr->next) {
    utf8 ns = (attr->ns == APR_XML_NS_NONE) ? NULL : ns_names[attr->ns];
    genxAddAttributeLiteral(gw, ns, (utf8) attr->name, (utf8) attr->value);
  }

  for (text = e->first_cdata.first; text; text = text->next)
    genxAddText(gw, (utf8) text->text);

  for (child = e->first_child; child; child = child->next) {
    char * problem;
    if ((problem = write_element(doc, child, gw)))
      return problem;
  }

  genxEndElement(gw);

  for (text = e->following_cdata.first; text; text = text->next)
    genxAddText(gw, (utf8) text->text);
  
  return NULL;
}

/* persist an XML document */
static char * write_entry(request_rec * r, apr_xml_doc * doc, char * filename)
{
  apr_status_t status;
  apr_file_t * fp;
  char * problem;
  genxWriter gw;
  genxElement * elements;
  genxAttribute * attributes;
  genxNamespace * namespaces;

  status = apr_file_open(&fp, filename,
			 APR_FOPEN_WRITE|APR_FOPEN_CREATE, PERMS, r->pool);
  if (status != APR_SUCCESS)
    return apr_psprintf(r->pool, "Failed to open '%s', status: %d",
			filename, status);

  problem = atom_genx_file_writer(r->pool, fp, &gw, NULL,
				  &elements, &attributes, &namespaces);
  if (problem)
    return apr_psprintf(r->pool, "Can't start XML writer: '%s'", problem);

  problem = write_element(doc, doc->root, gw);
  genxEndDocument(gw);
  apr_file_close(fp);
  if (problem)
    return problem;

  return NULL;
}

/* Here's how we try to create new files without locking anything.
 *  We build a name from the title (or slug, if provided) and try to
 *  create that.  If there's a collision, we append a large random
 *  number to the name, and try again, twice.  If it still fails, give up.
 * When this returns, there's a window during which the file being created
 *  exists, but is empty.  Hmm.
 */
/* this is in a separate routine to make unit-testing easier */
static apr_status_t entry_names(request_rec * r, char * dir, entry_t * entry)
{
  char * slug = (char *) apr_table_get(r->headers_in, "Slug");
  char * filename = NULL;
  char * path;
  apr_file_t * fp;
  apr_status_t status;
  char * html_name;
  char * title = NULL;
  apr_xml_elem * child;
  char * * namespaces = (char * *) entry->xml->namespaces->elts;

  /* have to fish the title out of the DOM to help create the filename */
  for (child = entry->xml->root->first_child; child; child = child->next) {
    if (strcmp(namespaces[child->ns], ATOM_NS) == 0 &&
	strcmp(child->name, "title") == 0) {
      title = (char *) child->first_cdata.first->text;
      break;
    }
  }
  if (title == NULL) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "POSTed entry has no atom:title element");
    return HTTP_FORBIDDEN;
  }

  if (!slug)
    slug = title;
  slug = decode_slug(slug, r->pool);

  filename = atom_join(r->pool, dir, slug, NULL);
  status = apr_file_open(&fp, filename, 
			 APR_FOPEN_WRITE|APR_FOPEN_CREATE|APR_FOPEN_EXCL,
			 PERMS, r->pool);
  if (status != APR_SUCCESS) {
    int i;
    for (i = 0; status != APR_SUCCESS && i < MAX_FILENAME_VARIATIONS; i++) {
      filename = apr_psprintf(r->pool, "%s-%ld",
			      atom_join(r->pool, dir, slug, NULL), random());
      status = apr_file_open(&fp, filename,
			     APR_FOPEN_WRITE|APR_FOPEN_CREATE|APR_FOPEN_EXCL,
			     PERMS, r->pool);
    }
  }
  if (status == APR_SUCCESS)
    apr_file_close(fp);
  else {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't create file starting with '%s/%s'", dir, slug);
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  filename = is_prefix(filename, dir) + 1; /* just the file part */

  path = is_prefix(dir, entry->pub->dir);

  /* path starts with /atom/, we want /pub/ instead */
  html_name = path + 1 + strlen(ATOM_DIR);
  html_name = atom_join(r->pool, 
			entry->pub->path, PUB_DIR, html_name, filename, NULL);
  path = atom_join(r->pool, entry->pub->path, path, filename, NULL);
  if (r->parsed_uri.port_str != NULL)
    path = apr_psprintf(r->pool, "http://%s:%s%s", r->hostname,
			r->parsed_uri.port_str, path);
  else
    path = apr_psprintf(r->pool, "http://%s%s", r->hostname, path);

  entry->filename = atom_join(r->pool, dir, filename, NULL);
  entry->html_name = html_name;
  entry->uri = path;

  entry->coll_base = atom_join(r->pool, entry->pub->dir,
			       is_prefix(r->uri, entry->pub->path), NULL);
  *(rindex(entry->coll_base, '/') + 1) = 0;

  return OK;
}

/* turn a string into a file; should be called "frans" or "un_snarf" */
static char * to_file(char * filename, char * data, apr_pool_t * pool)
{
  apr_status_t status;
  apr_file_t * fp;
  apr_size_t written;

  status = apr_file_open(&fp, filename, APR_FOPEN_WRITE|APR_FOPEN_CREATE,
			 PERMS, pool);
  if (status != APR_SUCCESS)
    return apr_psprintf(pool, "to_file: can't create '%s'", filename);

  status = apr_file_write_full(fp, (void *) data, strlen(data), &written);
  if (status != APR_SUCCESS)
    return apr_psprintf(pool, "to_file: can't write to tempfile '%s'",
			filename);
  apr_file_close(fp);
  return NULL;
}

/* turn a file into a string */
char * atom_snarf(char * filename, char * *  ret, apr_pool_t * pool) 
{
  apr_status_t status;
  apr_file_t * fp;
  apr_finfo_t finfo;
  char * buf;
  apr_size_t read_size;

  status = apr_stat(&finfo, filename, APR_FINFO_SIZE, pool);
  if (status != APR_SUCCESS)
    return apr_psprintf(pool, "Snarf: Can't stat '%s'", filename);
  
  status = apr_file_open(&fp, filename, APR_FOPEN_READ, PERMS, pool);
  if (status != APR_SUCCESS)
    return apr_psprintf(pool, "Snarf: Can't open '%s'", filename);

  read_size = finfo.size + 1;
  buf = apr_palloc(pool, read_size);
  if (buf == NULL)
    return apr_psprintf(pool, "Snarf: Can't allocate %ld bytes to read '%s'",
			(long) read_size, filename);
  status = apr_file_read(fp, buf, &read_size);
  if (status != APR_SUCCESS)
    return apr_psprintf(pool, "Snarf: Can't read %ld bytes from '%s'",
			(long) finfo.size, filename);
  buf[read_size] = 0;
  *ret = buf;
  return NULL;
}

static char * store_incoming_bits(request_rec * r, char * filename, int * size)
{
  apr_file_t * fp;
  apr_status_t status;
  apr_bucket_brigade * brigade;
  int seen_eos;
  apr_size_t total_read;
  apr_off_t max;

  status = apr_file_open(&fp, filename, APR_FOPEN_WRITE|APR_FOPEN_CREATE,
			 PERMS, r->pool);
  if (status != APR_SUCCESS)
    return apr_psprintf(r->pool, "Can't open file '%s'", filename);

  max = ap_get_limit_req_body(r);
  brigade = apr_brigade_create(r->pool, r->connection->bucket_alloc);

  seen_eos = 0;
  total_read = 0;

  do {
    apr_bucket * bucket;

    status = ap_get_brigade(r->input_filters, brigade, AP_MODE_READBYTES,
			    APR_BLOCK_READ, BUFSIZ);
    if (status != APR_SUCCESS)
      return apr_psprintf(r->pool, "Can't get_brigade");

    for (bucket = APR_BRIGADE_FIRST(brigade);
	 bucket != APR_BRIGADE_SENTINEL(brigade);
	 bucket = APR_BUCKET_NEXT(bucket)) {

      const char * data;
      apr_size_t len;

      if (APR_BUCKET_IS_EOS(bucket)) {
	seen_eos = 1;
	break;
      }
      if (APR_BUCKET_IS_METADATA(bucket))
	continue;
      status = apr_bucket_read(bucket, &data, &len, APR_BLOCK_READ);
      if (status != APR_SUCCESS)
	return apr_psprintf(r->pool, "Can't bucket_read");

      total_read += len;
      if (max && (total_read > max))
	/* XXX when errors are redone, this
	 *  should return HTTP_REQUEST_ENTITY_TOO_LARGE */
	return apr_psprintf(r->pool, "POSTed body exceed maximum size of %ld",
			    (long) max);

      status = apr_file_write_full(fp, (void *) data, len, &len);
      if (status != APR_SUCCESS)
	return apr_psprintf(r->pool, "Can't write to file '%s'", filename);
    }
    apr_brigade_cleanup(brigade);
  } while (!seen_eos);

  apr_brigade_destroy(brigade);
  apr_file_close(fp);
  *size = total_read;
  return NULL;
}

/* replace an existing media object */
static int replace_media(request_rec * r, media_t * media) 
{
  char * path;
  char * tempname;
  char * problem;
  pub_t * pub = media->mle->pub;
  char * media_type = (char *) apr_table_get(r->headers_in, "Content-Type");
  char * filename;

  path = atom_join(r->pool, pub->dir, media->bits_path, NULL);
  tempname = apr_psprintf(r->pool, "%s/%s%ld-%ld",
			  path, TEMP_PREFIX, random(), random());
  if ((problem = store_incoming_bits(r, tempname, &(media->size)))) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't same incoming POSTed bits to '%s': %s",
		 tempname, problem);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  filename = atom_join(r->pool, path, media->filename, NULL);
  if (apr_file_rename(tempname, filename, r->pool) != APR_SUCCESS) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't rename '%s' to '%s", tempname, filename);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  /* now save the media type */
  filename = apr_psprintf(r->pool, "%s.%s", filename, MEDIA_TYPE_SUFFIX);
  to_file(filename, media_type, r->pool);

  return OK;
}

/* save an incoming media object, updating the media struct */
static int save_new_media(request_rec * r, media_t * media) 
{
  char * dir;
  char * yyyymmdd = yyyymmdd_at(r->request_time, r->pool);
  char * problem;
  char * slug = (char *) apr_table_get(r->headers_in, "Slug");
  char * filename;
  char * media_type = (char *) apr_table_get(r->headers_in, "Content-Type");
  char * tempname;
  char * bdir, * mdir;
  pub_t * pub = media->mle->pub;

  /* uri for mle: collection (POSTED) uri minus /collection + 
   *  /yyyy/mm/dd/ + filename
   * uri for bits: collection uri - /collection 
   *  + /bits + /yyyy/mm/dd/ + filename
   * uri for media link entry: same, only without the /bits
   */
  dir = apr_pstrdup(r->pool, is_prefix(r->uri, pub->path)) + 1;
  *(rindex(dir, '/')) = 0; /* lost /collection$ */
  media->bits_path = apr_psprintf(r->pool, "%s/%s/%s", dir, BITS_DIR, yyyymmdd);
  media->mle_path = apr_psprintf(r->pool, "%s/%s", dir, yyyymmdd);
  bdir = apr_psprintf(r->pool, "%s/%s", pub->dir, media->bits_path);
  mdir = apr_psprintf(r->pool, "%s/%s", pub->dir, media->mle_path);
  if ((problem = atom_dir(bdir, r->pool)) ||
      (problem = atom_dir(mdir, r->pool))) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "save_new_media Can't access directory '%s': %s", dir,
		 problem);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  if (slug) 
    filename = apr_psprintf(r->pool, "%s-%ld-%ld", decode_slug(slug, r->pool),
			    random(), random());
  else 
    filename = apr_psprintf(r->pool, "a-%ld-%ld", random(), random());
  media->filename = apr_pstrdup(r->pool, filename);

  tempname = apr_psprintf(r->pool, "%s/%s/%s-%ld-%ld", pub->dir,
			  media->bits_path, TEMP_PREFIX, random(), random());
  if ((problem = store_incoming_bits(r, tempname, &(media->size)))) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't same incoming POSTed bits to '%s': %s",
		 tempname, problem);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  filename = apr_psprintf(r->pool, "%s/%s/%s",
			  pub->dir, media->bits_path, filename);
  if (apr_file_rename(tempname, filename, r->pool) != APR_SUCCESS) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't rename '%s' to '%s", tempname, filename);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  /* now save the media type */
  to_file(apr_psprintf(r->pool, "%s.%s", filename, MEDIA_TYPE_SUFFIX),
	  media_type, r->pool);
  return OK;
}
		       
/* save an entry into a file, updating fields in the entry struct */
static int save_new_entry(request_rec * r, entry_t * entry)
{
  apr_status_t status;
  char * dir;
  char * yyyymmdd = yyyymmdd_at(r->request_time, r->pool);
  char * problem;
  char * tempname;

  /* The URI path is:
   *  (request->uri - trailing "/collection") + /yyyy/mm/dd/ + filename
   * The filename is: just replace the pub->uri part at the front with
   *  pub->path
   */
  dir = apr_pstrdup(r->pool, is_prefix(r->uri, entry->pub->path)) + 1;
  *(rindex(dir, '/')) = 0; /* lost /collection$ */
  dir = apr_psprintf(r->pool, "%s/%s/%s", entry->pub->dir, dir, yyyymmdd);

  /* dir now has the right directory, make sure it's there */
  if ((problem = atom_dir(dir, r->pool))) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "save_new_entry Can't access directory '%s': %s", dir,
		 problem);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  /* choose new entry filename */
  if ((status = entry_names(r, dir, entry)) != OK)
    return status;

  /* one time only per entry */
  entry->id = new_id(r->pool);

  /* clean up entry */
  if ((problem = prepare_entry(r, entry)) != NULL) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Problem preparing posted entry: %s", problem);
    return HTTP_FORBIDDEN;
  }

  /* blast it into a tempfile */
  tempname = apr_psprintf(r->pool, "%s/%s%ld-%ld",
			  dir, TEMP_PREFIX, random(), random());
  if ((problem = write_entry(r, entry->xml, tempname))) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't write posted entry into '%s': %s", tempname, problem);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  if (apr_file_rename(tempname, entry->filename, r->pool) != APR_SUCCESS) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
		 "Can't rename from '%s' to '%s'", tempname, entry->filename);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  return OK;
}

/* it's in %-encoded UTF8.
 *  we'll turn it into UTF-8, then we'll turn anything that isn't an
 *  XML name character, or is a dot, into '-'
 */
static char * decode_slug(char * slug, apr_pool_t * pool) 
{
  utf8 decoded = (utf8) apr_pstrdup(pool, slug);
  utf8 cleaned;
  int in = 0, out = 0;
  int character;
  constUtf8 in_u;
  utf8 out_u;
  int dashed;

  while (slug[in]) {
    if (slug[in] != '%') {
      decoded[out++] = slug[in++];
      continue;
    }
    in++;
    if (!(apr_isxdigit(slug[in]) && apr_isxdigit(slug[in + 1])))
      return NULL;
    sscanf(slug + in, "%02x", &character);
    decoded[out++] = character;
    in += 2;
  }
  decoded[out] = 0;

  /* decoded now contains a pure UTF-8 version of the string */
  cleaned = (utf8) apr_pstrdup(pool, (char *) decoded);

  /* squish out illegal UTF-8 & non-XML chars.  Sam Ruby would say this
   *  is silent data loss, but that's OK, we give no guarantees aside from
   *  best-effort to create a reasonable URI
   */
  genxScrubText(char_util, (constUtf8) decoded, cleaned);
  if (*cleaned == 0)
    return NULL;

  /* now we'll walk through and move characters over, replacing everything
   *  that's not a NameChar with '-'
   */
  decoded = cleaned;
  cleaned = (utf8) apr_pstrdup(pool, (char *) decoded);
  in_u = decoded;
  out_u = cleaned;
  dashed = 1; /* suppress leading dashes */
  while (*in_u) {
    utf8 old_in = (utf8) in_u;
    int next_c = genxNextUnicodeChar(&in_u);
    if ((next_c != '.') &&
	(genxCharClass(char_util, next_c) & GENX_NAMECHAR)) {
      while (old_in < in_u)
	*out_u++ = *old_in++;
      dashed = 0;
    }
    else {
      /* suppress multiple dashes */
      if (!dashed)
	*out_u++ = '-';
      dashed = 1;
    }
  }
  *out_u = 0;

  /* now we have to %-encode again */
  decoded = cleaned;
  for (in = out = 0; decoded[in]; in++)
    if (decoded[in] < 128)
      out++;
    else
      out += 3;
  cleaned = (utf8) apr_palloc(pool, out + 1);
  for (in = out = 0; decoded[in]; in++) {
    if (decoded[in] == '%') {
      cleaned[out++] = '%';
      cleaned[out++] = '%';
    }
    if (decoded[in] < 128)
      cleaned[out++] = decoded[in];
    else {
      char s[4];
      sprintf(s, "%%%02X", decoded[in]);
      cleaned[out++] = s[0];
      cleaned[out++] = s[1];
      cleaned[out++] = s[2];
    }
  }
  cleaned[out] = 0;

  return (char *) cleaned;
}

/* Create the APP service doc; return NULL or an error message */
static char * write_service_doc(pub_t * pub, apr_pool_t * pool) 
{
  apr_file_t * service;
  char * problem;
  apr_status_t status;
  genxElement * elements;
  genxAttribute * attributes;
  genxNamespace * namespaces;
  genxWriter gw;
  char * service_file;
  char * val;

  service_file = apr_psprintf(pool, "%s/%s/%s",
			      pub->dir, ATOM_DIR, SERVICE_FILE);

  /* we can do the following because we're only called at config time */
  apr_file_remove(service_file, pool);
  status = apr_file_open(&service, service_file,
			 APR_FOPEN_WRITE|APR_FOPEN_CREATE, PERMS, pool);
  if (status != APR_SUCCESS)
    return apr_psprintf(pool, "Failed to open '%s', status: %d",
			service_file, status);

  problem = atom_genx_file_writer(pool, service, &gw, NULL,
				  &elements, &attributes, &namespaces);
  if (problem)
    return apr_psprintf(pool, "Can't start XML writer: '%s'", problem);

  genxStartElement(elements[APP_SERVICE_E]);
  genxAddNamespace(namespaces[APP_NS_NUM], NULL);
  genxAddNamespace(namespaces[ATOM_NS_NUM], NULL);
			      
  genxAddCharacter(gw, '\n');
  genxStartElement(elements[APP_WORKSPACE_E]);
  atom_genx_simple_element(gw, elements[ATOM_TITLE_E], pub->title);

  genxStartElement(elements[APP_COLLECTION_E]);
  val = apr_psprintf(pool, "%s/%s/%s/%s/%s", pub->path, ATOM_DIR,
		     ENTRIES_DIR, ENTRIES_COLL, COLLECTION_FILE);
  genxAddAttribute(attributes[HREF_A], (utf8) val);
  genxAddCharacter(gw, '\n');
  atom_genx_simple_element(gw, elements[ATOM_TITLE_E], "Entries");
  genxEndElement(gw); /* collection */
  genxAddCharacter(gw, '\n');

  genxStartElement(elements[APP_COLLECTION_E]);
  val = apr_psprintf(pool, "%s/%s/%s/%s/%s", pub->path, ATOM_DIR,
		     MEDIA_DIR, MEDIA_COLL, COLLECTION_FILE);
  genxAddAttribute(attributes[HREF_A], (utf8) val);
  genxAddCharacter(gw, '\n');
  atom_genx_simple_element(gw, elements[ATOM_TITLE_E], "Media");
  atom_genx_simple_element(gw, elements[APP_ACCEPT_E], "*/*");
  genxEndElement(gw); /* collection */
  genxAddCharacter(gw, '\n');

  genxEndElement(gw); /* workspace */
  genxEndElement(gw); /* service */
  genxEndDocument(gw);
  
  apr_file_close(service);
  return NULL;
}

/* got tired of calling apr_stat */
int atom_file_exists(char * name, apr_pool_t * pool) 
{
  apr_finfo_t finfo;
  return (apr_stat(&finfo, name, APR_FINFO_MIN, pool) == APR_SUCCESS);
}

/* probe to see if a pub exists; create it if not.  Return an error message
 *  if unsuccessful. It might be better to have an external
 *  make-a-pub tool, but then I'd have to manage it & keep track of it.
 */
static char * build_pub(pub_t * pub, server_rec * s,
			apr_pool_t * pperm, apr_pool_t * ptemp)
{
  char * re_s;
  static char * yyyymmdd = "[0-9][0-9][0-9][0-9]/[0-9][0-9]/[0-9][0-9]";
  char * problem;
  char * dirname;
  apr_dir_t * dir;
  apr_finfo_t finfo;

  /* We need to ensure that (minimally) the PUB and ATOM subdirs exist,
   *  and are writeable, and that the ENTRIES and MEDIA collections exist.
   */
  if ((problem = atom_dir(pub->dir, ptemp)))
    return problem;
  
  if ((problem = atom_dir(atom_join(ptemp, pub->dir, PUB_DIR, NULL), ptemp)))
    return problem;
  if ((problem = atom_dir(atom_join(ptemp, pub->dir, ATOM_DIR, NULL), ptemp)))
    return problem;

  re_s = atom_join(ptemp, pub->dir, PUB_DIR, ENTRIES_COLL, NULL);
  if ((problem = atom_dir(re_s, ptemp)))
    return problem;
  re_s = atom_join(ptemp, pub->dir, PUB_DIR, MEDIA_COLL, NULL);
  if ((problem = atom_dir(re_s, ptemp)))
    return problem;
  re_s = atom_join(ptemp, pub->dir, ATOM_DIR, ENTRIES_DIR, ENTRIES_COLL, NULL);
  if ((problem = atom_dir(re_s, ptemp)))
    return problem;
  re_s = atom_join(ptemp, pub->dir, ATOM_DIR, MEDIA_DIR, MEDIA_COLL, NULL);
  if ((problem = atom_dir(re_s, ptemp)))
    return problem;

  /* pre-compute a bunch of strings for query-time uri analysis:
   * - a string to match the service doc */
  pub->service_doc = atom_join(pperm, pub->path, ATOM_DIR, SERVICE_FILE, NULL);

  /* - a regex to detect html content */
  re_s = atom_join(ptemp, pub->path, PUB_DIR, NULL);
  re_s = apr_psprintf(ptemp, "^%s/(index\\.html)|(%s/[^/]*\\.x?html)",
		      re_s, yyyymmdd);
  pub->html_re = ap_pregcomp(pperm, re_s, AP_REG_EXTENDED);

  /* - a regex to detect atom feeds */
  re_s = atom_join(ptemp, pub->path, PUB_DIR, FEED_FILE, NULL);
  re_s = apr_psprintf(ptemp, "^%s", re_s);
  pub->feed_re = ap_pregcomp(pperm, re_s, AP_REG_EXTENDED);

  /* - a regex to detect atom entries */
  re_s = atom_join(ptemp, pub->path, ATOM_DIR, NULL);
  re_s = apr_psprintf(ptemp, "^%s/[^/]*/[^/]*/%s/[^.][^.]*", re_s, yyyymmdd);
  pub->entry_re = ap_pregcomp(pperm, re_s, AP_REG_EXTENDED);

  /* - a regex to match a collection */
  re_s = atom_join(ptemp, pub->path, ATOM_DIR, NULL);
  re_s = apr_psprintf(ptemp, "^%s/[^/]*/[^/]*/%s", re_s, COLLECTION_FILE);
  pub->collection_re = ap_pregcomp(pperm, re_s, AP_REG_EXTENDED);

  /* - a regex to match a media object */
  re_s = atom_join(ptemp, pub->path, ATOM_DIR, MEDIA_DIR, NULL);
  re_s = apr_psprintf(ptemp, "^%s/[^/]*/%s/%s/[^.][^.]*",
		      re_s, BITS_DIR, yyyymmdd);
  pub->media_re = ap_pregcomp(pperm, re_s, AP_REG_EXTENDED);

  /* - a prefix to match entry/media collections */
  pub->entries_prefix =
    atom_join(pperm, pub->path, ATOM_DIR, ENTRIES_DIR, NULL);
  pub->media_prefix = atom_join(pperm, pub->path, ATOM_DIR, MEDIA_DIR, NULL);

  /* - a regex to use in turning a media URI into the corresponding MLE URI */
  re_s = atom_join(ptemp, pub->path, ATOM_DIR, MEDIA_DIR, NULL);
  re_s = apr_psprintf(pperm, "^(%s/[^/]*)/%s(/.*)$", re_s, BITS_DIR);
  pub->mle_maker = ap_pregcomp(pperm, re_s, AP_REG_EXTENDED);
  
  /* Recreate: service doc, collection docs, index.html
   */
  if ((problem = write_service_doc(pub, ptemp)))
    return problem;

  /* any subdirectory of MEDIA_DIR or ENTRIES_DIR is a collection, so it 
   * needs a feed written
   */
  dirname = apr_psprintf(ptemp, "%s/%s/%s", pub->dir, ATOM_DIR, ENTRIES_DIR);

  /* this routine has already successfully visited this dir, thus the lack
   * of checking */
  apr_dir_open(&dir, dirname, ptemp);
  while (apr_dir_read(&finfo, APR_FINFO_NAME|APR_FINFO_TYPE, dir) == APR_SUCCESS) {
    if (finfo.filetype == APR_DIR && index(finfo.name, '.') == NULL) {
      char * coll = atom_join(ptemp, dirname, finfo.name, NULL);
      char * id_name = atom_join(ptemp, coll, ID_FILE, NULL);

      /* if there's no ID file (collection is being created) make one */
      if (!atom_file_exists(id_name, ptemp)) {
	apr_file_t * fp;
	char * id = new_id(ptemp);
	apr_size_t written;

	apr_file_open(&fp, id_name, APR_FOPEN_WRITE|APR_FOPEN_CREATE,
		      PERMS, ptemp);
	apr_file_write_full(fp, id, strlen(id), &written);
	apr_file_close(fp);
      }

      if ((problem = write_collection(ptemp, s, pub, coll,
				      DEFAULT_ENTRIES_IN_FEED))) {
	apr_dir_close(dir);
	return problem;
      }
    }
  }
  apr_dir_close(dir);

  dirname = atom_join(ptemp, pub->dir, ATOM_DIR, MEDIA_DIR, NULL);

  /* this routine has already successfully visited this dir, thus the
   * lack of checking */
  apr_dir_open(&dir, dirname, ptemp);
  while (apr_dir_read(&finfo, APR_FINFO_NAME|APR_FINFO_TYPE, dir) == APR_SUCCESS) {
    if (finfo.filetype == APR_DIR && index(finfo.name, '.') == NULL) {
      char * coll = atom_join(ptemp, dirname, finfo.name, NULL);
      char * id_name = atom_join(ptemp, coll, ID_FILE, NULL);

      /* if there's no ID file (collection is being created) make one */
      if (!atom_file_exists(id_name, ptemp)) {
	apr_file_t * fp;
	char * id = new_id(ptemp);
	apr_size_t written;

	apr_file_open(&fp, id_name, APR_FOPEN_WRITE|APR_FOPEN_CREATE,
		      PERMS, ptemp);
	apr_file_write_full(fp, id, strlen(id), &written);
	apr_file_close(fp);
      }

      if ((problem = write_collection(ptemp, s, pub, coll,
				      DEFAULT_ENTRIES_IN_FEED))) {
	apr_dir_close(dir);
	return problem;
      }
    }
  }
  apr_dir_close(dir);
  return NULL;
}

/* does name represent a writeable directory, which we may need to
 *  create as a side effect?
 */
static char * atom_dir(char * name, apr_pool_t * p) 
{
  apr_finfo_t finfo;
  char * tempname;
  apr_file_t * f;
  apr_status_t s = apr_stat(&finfo, name, APR_FINFO_TYPE, p);

  if (s == APR_SUCCESS) {
    if (finfo.filetype != APR_DIR) 
      return apr_psprintf(p, "'%s' is not a directory", name);
  }
  else {
    /* couldn't stat so let's try to make it */
    s = apr_dir_make_recursive(name, PERMS, p);
    if (s != APR_SUCCESS) 
      return apr_psprintf(p, "Can't create directory '%s'", name);
  }

  /* we want to see if the directory is writeable... we could screw
   *  around with the stupid uid_current and all that, or we could just
   *  try to open a file. 
   */
  tempname = apr_psprintf(p, "%s/%ld-%ld", name, random(), random());
  s = apr_file_open(&f, tempname,
		    APR_FOPEN_WRITE|APR_FOPEN_CREATE|APR_FOPEN_DELONCLOSE,
		    PERMS, p);
  if (s == APR_SUCCESS) {
    apr_file_close(f);
    return NULL;
  }
  else 
    return apr_psprintf(p, "Directory '%s' is not writeable", name);
  
}

/* Build index structure for looking up if a uri is in the URI space
 *  of one of our pubs.
 *
 * Here's the problem; we have to look at every URI to decide if
 *  if it falls into the URI space of one of our pubs.  This is easy
 *  because it's prefix-matching.  Unfortunately, if we have a lot of pubs,
 *  it'll be an unacceptable tax on the whole httpd.
 * One plan would be to have a big regex with
 *  ^(pub1|pub2|pub3|... )
 *  which would tell us yes-no but then we'd still have to run through
 *  all the pubs, and if this server is *mostly* going to be servicing
 *  our pubs, we'll end up with sequences of prefix checks.  You could get
 *  clever and do:
 *  ^((pub1)|(pub2)|... )
 *  then look for $2 and $3 and so on, but unfortunately AP_MAX_REG_MATCH
 *  is 10.
 *  So what we do, since we know that URI characters are only in the range
 *  0-127, is make a simple DFA with arrays of 128 ints.  This makes
 *  the check for being in the URI space a pretty-fast linear function of
 *  the unique prefix.  We use ints rather than chars because when you
 *  actually match, you need to signal this isn't another turn of the DFA
 *  crank, it's a match, and the rest of the bits give you an index into "pubs".
 * Each state in the DFA is an int[] indexed by the current character;
 *  entry value of 0 means fail-to-match, a positive integer is the index
 *  of the next state, a negative integer indicates a successful match.
 */
/* set up the index-building recursion */
static int build_index(apr_pool_t * pperm, apr_pool_t * ptemp) 
{
  int * state0;
  int * * slot0;
  int * mask0 = (int *) apr_palloc(ptemp, atom_pubs->nelts * sizeof(int));
  int i;
  int ret;

  pub_index = apr_array_make(pperm, 10, sizeof(int *));
  state0 = (int *) apr_pcalloc(pperm, 128 * sizeof(int));
  slot0 = (int * * ) apr_array_push(pub_index);
  *slot0 = state0;
  for (i = 0; i < atom_pubs->nelts; i++)
    mask0[i] = 1;
  ret = index_step(state0, mask0, 0, pperm, ptemp);
  /* XXX log the size of the DFA */
  return ret;
}

/* build one level of the pub-finding DFA (see big comments above).
 * The idea is that all the pubs in the subset of atom_pubs flagged in
 *  this_mask have a common prefix of length pref_length, so we're going
 *  to look at the next letter in each & figure out where the common-prefix
 *  runs out.
 * Return 0 if all went OK, 1 if error.
 */
static int index_step(int * this_state, int * this_mask, int prefix_length,
		      apr_pool_t * pperm, apr_pool_t * ptemp) 
{
  int * letter_counts = (int *) apr_pcalloc(ptemp, 128 * sizeof(int));
  pub_t * publist = (pub_t *) atom_pubs->elts;
  int pub;
  int letter;

  for (pub = 0; pub < atom_pubs->nelts; pub++)
    if (this_mask[pub])
      letter_counts[(int) (publist[pub].path[prefix_length])]++;

  /* more than one pub with the same base uri is a bug */
  if (letter_counts[0] > 1) {
    /* XXX TODO log an error */
    return 1;
  }

  for (letter = 0; letter < 128; letter++) {
    if (letter_counts[letter] == 0)
      continue;

    else if (letter_counts[letter] == 1) {
      for (pub = 0; pub < atom_pubs->nelts; pub++) {
	if (this_mask[pub] && (publist[pub].path[prefix_length] == letter)) {
	  this_state[letter] = -1 * (pub + 1); /* -1 * 0 == 0, sigh */
	  break;
	}
      }
    }
	  
    else {
      /* need to build another state in the machine */
      int * next_mask =
	(int *) apr_pcalloc(ptemp, atom_pubs->nelts * sizeof(int));
      int next_index = pub_index->nelts;
      int * * next_slot = (int * *) apr_array_push(pub_index);
      int * next_state = (int *) apr_pcalloc(pperm, 128 * sizeof(int));
      *next_slot = next_state;

      for (pub = 0; pub < atom_pubs->nelts; pub++) {
	if (this_mask[pub] && (publist[pub].path[prefix_length] == letter)) {
	  next_mask[pub] = 1;
	  this_state[letter] = next_index;
	}
      }
      if (index_step(next_state, next_mask, prefix_length + 1, pperm, ptemp))
	return 1;
    }
  }
  return 0;
}

/* is this URI one of ours? If so, return its pub_t structure, else NULL */
/* Hmm, this will blow chunks if someone manages to slip in a uri with
 *  non-ASCII characters.  They should be %-encoded by the time they get
 *  here, right?
 */
static pub_t * find_pub(char * uri_in) 
{
  pub_t * publist = (pub_t *) atom_pubs->elts;
  int * * states = (int * *) pub_index->elts;
  int * state = states[0];
  char * uri = uri_in;

  while (*uri) {
    int next = state[(int) (*uri++)];
    if (next == 0)
      return NULL;
    else if (next < 0) {
      pub_t * pub = publist + ((next * -1) - 1);
      return (is_prefix(uri_in, pub->path)) ? pub : NULL;
    }
    else
      state = states[next];
  }
  
  /* prefix was ambiguous, we ran out of uri with multiple possibilities
   *  still in play
   */
  /* XXX TODO log an error message */
  return NULL;
}

/* Discover whether 'prefix' is a prefix of s. Return NULL if not; if it is,
 *  return a pointer to the part of s following the prefix
 */
static char * is_prefix(char * s, char * prefix) 
{
  for ( ; *prefix && *s && (*prefix == *s); s++, prefix++)
    /* empty */ ;

  return (*prefix == 0) ? s : NULL;
}

static char * yyyymmdd_at(apr_time_t when, apr_pool_t * pool)
{
  apr_time_exp_t times;

  apr_time_exp_lt(&times, when);
  return apr_psprintf(pool, "%04d/%02d/%02d", times.tm_year + 1900,
		      times.tm_mon + 1, times.tm_mday);
}

/* The problem is that mod_atom is forever gluing together fragments to make
 *  pathnames of one kind or another.  It's tough to keep track of whether
 *  or not the fragments end with '/', particularly given that the AtomPub
 *  directive arguments might or might not have slashes on the end.
 * So this glues its arguments together, ensuring that there is exactly
 *  one connecting slash.
 */
char * atom_join(apr_pool_t * pool, ...)
{
  va_list va;
  char * path;
  char * joined;
  int buf_len = 0;
  int in, out;

  /* Figure out how much memory we'll need.  You could avoid this step, but
   *  it'd cost you a function call per character.  Alternatively, you could
   *  build the answer in a local char[24000] or something which would *never*
   *  fail but That Would Be Wrong.
   */
  va_start(va, pool);
  while ((path = va_arg(va, char *)))
    buf_len += strlen(path) + 1;
  joined = apr_palloc(pool, buf_len + 1);

  out = 0;
  va_start(va, pool);
  while ((path = va_arg(va, char *)))
  {
    in = (out != 0 && path[0] == '/') ? 1 : 0;
    while (path[in])
      joined[out++] = path[in++];

    if (in > 0 && path[in - 1] != '/')
      joined[out++] = '/';
  }

  if (out > 0 && joined[out - 1] == '/')
    out--;
  joined[out] = 0;
  return joined;
}


/* all this hookup-to-httpd grunge is explicitly excluded from unit testing */
#if !defined UNIT_TEST

/* register our interests */
static void hooks(apr_pool_t * p)
{
  ap_hook_pre_config(pre_config, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_post_config(post_config, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_handler(handler, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_type_checker(type_checker, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_map_to_storage(map_to_storage, NULL, NULL, APR_HOOK_MIDDLE);
}

/* just one directive */
static const command_rec directives[] = {
  AP_INIT_TAKE_ARGV("AtomPub", atompub_directive, NULL, RSRC_CONF,
		    "AtomPub uri-path real-dir title author"),
  {NULL}
};

module AP_MODULE_DECLARE_DATA atom_module = {
  STANDARD20_MODULE_STUFF,
  NULL,                       /* create per-directory config structure */
  NULL,                       /* merge per-directory config structures */
  NULL,                       /* create per-server config structure */
  NULL,                       /* merge per-server config structures */
  directives,                 /* handle config directive */
  hooks                       /* registrar */
};
#endif

#if defined(UNIT_TEST)

/* this has to be #included in so its code can call all the static
 *  routines in mod_atom.c
 */
#include "unit_test.c"

#endif

