/*
 * Copyright (c) 2007 Sun Microsystems, Inc. All rights reserved
 * Use is subject to license terms.
 *
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#if defined(UNIT_TEST)
#include "unit_test.h"
#else

#include "apr.h"
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_hash.h"
#include "apr_user.h"
#include "apr_lib.h"
#include "apr_optional.h"
#include "apr_xml.h"
#include "apr_uuid.h"

#define APR_WANT_STRFUNC
#define APR_WANT_MEMFUNC
#include "apr_want.h"

#include "ap_config.h"
#include "util_xml.h"
#include "util_filter.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_request.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_main.h"
#include "util_script.h"
#include "http_core.h"
#include "mod_include.h"

#endif

#define ATOM_DIR "atom"
#define BITS_DIR "bits"
#define COLLECTION_FILE "collection"
#define ENTRIES_COLL "entries"
#define ENTRIES_DIR "e"
#define FEED_FILE "feed"
#define FRONT_PAGE "index.html"
#define LINK_PREFIX "L."
#define ID_FILE "id."
#define MEDIA_COLL "media"
#define MEDIA_DIR "m"
#define MEDIA_TYPE_SUFFIX "mt"
#define PUB_DIR "pub"
#define SERVICE_FILE "service"
#define TEMP_PREFIX "T."

#define WEAK_PREFIX "W/"

#define PERMS APR_FPROT_UREAD|APR_FPROT_UWRITE|APR_FPROT_UEXECUTE|APR_FPROT_GREAD|APR_FPROT_GWRITE|APR_FPROT_GEXECUTE

#define MAX_FILENAME_VARIATIONS 2

#define DEFAULT_ENTRIES_IN_FEED 20

#define ATOM_NS "http://www.w3.org/2005/Atom"
#define APP_NS "http://www.w3.org/2007/app"
#define XHTML_NS "http://www.w3.org/1999/xhtml"
#define XML_NS "http://www.w3.org/XML/1998/namespace"

/* predefined elements */
#define APP_ACCEPT_E 0
#define APP_CATEGORIES_E (APP_ACCEPT_E + 1)
#define APP_COLLECTION_E (APP_CATEGORIES_E + 1)
#define APP_EDITED_E (APP_COLLECTION_E + 1)
#define APP_SERVICE_E (APP_EDITED_E + 1)
#define APP_WORKSPACE_E (APP_SERVICE_E + 1)

#define ATOM_AUTHOR_E (APP_WORKSPACE_E + 1)
#define ATOM_CATEGORY_E (ATOM_AUTHOR_E + 1)
#define ATOM_CONTENT_E (ATOM_CATEGORY_E + 1)
#define ATOM_CONTRIBUTOR_E (ATOM_CONTENT_E + 1)
#define ATOM_EMAIL_E (ATOM_CONTRIBUTOR_E + 1)
#define ATOM_ENTRY_E (ATOM_EMAIL_E + 1)
#define ATOM_FEED_E (ATOM_ENTRY_E + 1)
#define ATOM_GENERATOR_E (ATOM_FEED_E + 1)
#define ATOM_ICON_E (ATOM_GENERATOR_E + 1)
#define ATOM_ID_E (ATOM_ICON_E + 1)
#define ATOM_LINK_E (ATOM_ID_E + 1)
#define ATOM_LOGO_E (ATOM_LINK_E + 1)
#define ATOM_NAME_E (ATOM_LOGO_E + 1)
#define ATOM_PUBLISHED_E (ATOM_NAME_E + 1)
#define ATOM_RIGHTS_E (ATOM_PUBLISHED_E + 1)
#define ATOM_SOURCE_E (ATOM_RIGHTS_E + 1)
#define ATOM_SUBTITLE_E (ATOM_SOURCE_E + 1)
#define ATOM_SUMMARY_E (ATOM_SUBTITLE_E + 1)
#define ATOM_TITLE_E (ATOM_SUMMARY_E + 1)
#define ATOM_UPDATED_E (ATOM_TITLE_E + 1)
#define ATOM_URI_E (ATOM_UPDATED_E + 1)

#define ATOM_ELEMENT_MAX (ATOM_URI_E + 1)

/* predefined attributes */
#define FIXED_A 0
#define HREFLANG_A (FIXED_A + 1)
#define HREF_A (HREFLANG_A + 1)
#define LABEL_A (HREF_A + 1)
#define LENGTH_A (LABEL_A + 1)
#define REL_A (LENGTH_A + 1)
#define SCHEME_A (REL_A + 1)
#define SRC_A (SCHEME_A + 1)
#define TERM_A (SRC_A + 1)
#define TITLE_A (TERM_A + 1)
#define TYPE_A (TITLE_A + 1)
#define URI_A (TYPE_A + 1)
#define VERSION_A (URI_A + 1)

/* predefined namespaces */
#define ATOM_NS_NUM 0
#define APP_NS_NUM (ATOM_NS_NUM + 1)
#define XHTML_NS_NUM (APP_NS_NUM + 1)
#define NS_NUM_MAX (XHTML_NS_NUM + 1)

#define ATOM_ATTRIBUTE_MAX (VERSION_A + 1)

#include "genx.h"

/* oh dear, globals polluting httpd's namespace :( */
char * atom_join(apr_pool_t * pool, ...);
genxStatus atom_genx_simple_element(genxWriter gw, genxElement e,
				    char * content);
char * atom_genx_file_writer(apr_pool_t * p, apr_file_t * f,
			     genxWriter * gwp,
			     genxSender * * gsp,
			     genxElement * * ep,
			     genxAttribute * * ap,
			     genxNamespace * * np);
char * atom_genx_client_writer(apr_pool_t * p, request_rec * r,
			       genxWriter * gwp,
			       genxSender * * gsp,
			       genxElement * * ep,
			       genxAttribute * * ap,
			       genxNamespace * * np);

char * atom_iter_new(apr_pool_t * pool, char * dir, void * * iter_p);
char * atom_iter_next(void * iter, char * * next_p);

char * atom_snarf(char * filename, char * *  ret, apr_pool_t * pool);
int atom_file_exists(char * name, apr_pool_t * pool);
